// Copyright 2019 autocareai.com. All rights reserved.
// Use of this source code is governed by autocareai.com.

package zr

import (
	"fmt"
	"gd_vehicle/apis"
	"gd_vehicle/common.in/cache"
	"gd_vehicle/common.in/utils"
	"gd_vehicle/consts"
	"gd_vehicle/errors"
	"gd_vehicle/thirdparty"
	dutils "gd_vehicle/utils"
	"github.com/tidwall/gjson"
	"go.uber.org/zap"
)

func getZrToken(hostPath string, tryAgain bool, httpTimeout int) (string, error) {
	token := ""
	redisKey := "zr:token"
	if !tryAgain {
		token, _ = cache.Redis.Get(redisKey)
		if token != "" {
			return token, nil
		}
	}

	resp, err := thirdparty.ZrLogin(hostPath,httpTimeout)
	if err != nil {
		l.Error("thridPart",
			zap.String("call", "ZrLogin"),
			zap.String("args", ""),
			zap.String("error", err.Error()))
		return "", err
	}

	code := gjson.GetBytes(resp, "code").String()
	if code != "0" {
		// 失败的状态
		msg := gjson.GetBytes(resp, "msg").String()
		return string(resp), fmt.Errorf("%s",msg)
	}


	token = gjson.GetBytes(resp, "token").String()
	_, _ = cache.Redis.SetEx(redisKey, 3500, token)

	// 写入请求日志
	return token, nil
}

/*func GetZrTwoElementVerifyGuangdongData(rdata string,respMap map[string]string){
	if rdata == "1"{
		respMap["is_match"] = rdata
	}else if rdata == "2"{
		respMap["is_match"] = "-1"
	}
}*/

func getZrGuangdongKey(req *apis.ThirdpartRequest) string {
	key := consts.ZRTwoElementVerifyPre + req.Params["plateNo"] +  req.Params["plateType"] + req.Params["name"]
	return key
}


func getZrGuangdongPlateNoKey(req *apis.ThirdpartRequest) string {
	key := consts.ZRTwoElementVerifyPre + req.Params["plateNo"] +  req.Params["plateType"]
	return key
}

func paramZrGuangdongPlateNoData(ret string) (reply string, err error){
	return ret,nil
}


func paramZrGuangdongData(ret string) (reply string, err error) {
	code := gjson.Get(ret, "code").String()
	result := gjson.Get(ret, "data.msg").String()
	if code == "0" {
		if result == "OK"{
			reply = "1"
		}else if result == "NOT OK"{
			reply = "-1"
		}
		return reply,nil
	} else if code == "10020"{
		return "", errors.NoRecord
	}else if code == "10003"{
		return "",errors.UserTokenExpire
	}

	return "", errors.VendorError
}

func TwoElementVerifyGuangdongLocal(req *apis.ThirdpartRequest) (reply string, err error) {
	name := req.Params["name"]
	localOwner, err := dutils.GetDataByRedis(getZrGuangdongPlateNoKey(req), int64(req.Req.ReuseTime), paramZrGuangdongPlateNoData)
	if err == nil {
		if localOwner == name{
			return "1",nil
		}else {
			return "-1",nil
		}
	}

	reply, err = dutils.GetDataByRedis(getZrGuangdongKey(req), int64(req.Req.ReuseTime), paramZrGuangdongData)
	return reply, err
}

func TwoElementVerifyGuangdong(req *apis.ThirdpartRequest) (reply string, err error) {
	plateNo := req.Params["plateNo"]
	name := req.Params["name"]
	if plateNo == "" || name == "" {
		return "",errors.ArgsError
	}



	req.LReq.Search = plateNo
	token,err := getZrToken(req.ProviderInfo.ThirdpartHost,req.TryAgain,req.ProviderInfo.Timeout)
	if err != nil{
		l.Error("thridPart",
			zap.String("call", "getZrToken"),
			zap.String("args",utils.MarshalJsonString(req.Params)),
			zap.String("error", err.Error()))
		req.LReq.ResponseParams = err.Error()
		dutils.SetAccessLogReqCodeMsgState(req.LReq, 1101, "gd_1101", err.Error(), false)
		return "",errors.VendorError
	}

	ret, err := thirdparty.ZrHttpClient(req.ProviderInfo.ThirdpartHost+"/"+req.ProviderInfo.ThirdpartApiRouter, token,req.Params,req.ProviderInfo.Timeout)
	if err != nil {
		l.Error("func",
			zap.String("call", "ZrHttpClient"),
			zap.String("args", utils.MarshalJsonString(req.Params)),
			zap.String("error", err.Error()))
		req.LReq.ResponseParams = err.Error()
		dutils.SetAccessLogReqCodeMsgState(req.LReq, 1101, "gd_1101", "三方错误,请求失败", false)
		return "", errors.VendorError
	} else {
		req.LReq.ResponseParams = string(ret)
		reply ,err = paramZrGuangdongData(string(ret))
		message := gjson.GetBytes(ret, "msg").String()
		code := gjson.GetBytes(ret, "code").String()
		if err == nil {
			if reply == "-1"{
				// 不一致，不计费
				dutils.SetAccessLogReqCodeMsgState(req.LReq, 1100, "gd_1100", message, false)
				dutils.StoreDataToRedis(getZrGuangdongKey(req), string(ret))
			}else {
				// 一致
				dutils.SetAccessLogReqCodeMsgState(req.LReq, 0, code, message, true)
				dutils.StoreDataToRedis(getZrGuangdongPlateNoKey(req), req.Params["name"])
			}

			//dutils.StoreDataToRedis(getZrGuangdongKey(req), string(ret))
		} else if err == errors.NoRecord {
			dutils.SetAccessLogReqCodeMsgState(req.LReq, 1100, code, message, false)
			dutils.StoreDataToRedis(getZrGuangdongKey(req), string(ret))
		}else if err == errors.UserTokenExpire{
			/*req.Times = req.Times + 1
			if req.Times >= 3 {
				return "", errors.VendorError
			}*/
			if !req.TryAgain {
				req.TryAgain = true
				return TwoElementVerifyGuangdong(req)
			}
			dutils.SetAccessLogReqCodeMsgState(req.LReq, 1101, code, message, false)
		} else {
			l.Error("func",
				zap.String("call", "ZrHttpClient"),
				zap.String("args", utils.MarshalJsonString(req.Params)),
				zap.String("error", err.Error()))
			//codeInt,_ := strconv.Atoi(code)
			dutils.SetAccessLogReqCodeMsgState(req.LReq, 1101, code, message, false)
		}
		return reply, err
	}
}

